/* html.c
   Module for parsing HTML buffers */

#include "config.h"
#include <ctype.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include "geturl.h"
#include "html.h"

/* Match a string against a null-terminated list of identifiers. */
int
idmatch(char **list, char *what)
{
   int i;

   for (i = 0; list[i]; i++)
      if (!strcasecmp(list[i], what))
	 return 1;
   return 0;
}

/* The following function parses an HTML buffer buf, of size
   bufsize searching for tags. When a tag is encountered, the
   routine will extract its components. If "src", "href" or
   "background" are found, their address is returned, and so
   is the length of the string. Returns NULL if no URL
   is found.

   You must also provide an id for the specified buffer, so
   that the function can associate it with a parser state.
   Thus several buffers can be parsed, and all the switching
   between them will be accurate, provided each of them has
   a unique ID (e.g. its filename).

   Latest: add no-quote support (e.g. href=http://something/)

   Note: possible bugs at the end of buffers, or with small
   buffers */

char *
htmlfindurl(char *buf, int bufsize, int *size, char *id)
{
   char *ptr;
   state_t *s;
   /* NULL-terminated list of modifiers someone
      would want to follow -- feel free to edit to
      suite your needs: */
   static char *allow_mod[] = {
      "href",
      "src",
      "background",
      NULL
   };

   s = getstate(id);
   if (s->in_quote)
   {
      for (; *buf != '"' && bufsize > 0; buf++, bufsize--);
      s->in_quote = 0;
      if (*buf != '"')
      {
	 if (s->idname)
	    free(s->idname);
	 return NULL;
      }
      ++buf; --bufsize;
   }
   for (ptr = NULL; bufsize > 0; bufsize--, buf++)
   {
      switch (*buf)
      {
	 /* Since we're bound to be outside a quote,
	    '<' means entering a tag */
	 case '<':
	    s->in_tag = 1;
	    if (s->idname)
	    {
	       free(s->idname);
	       s->idname = NULL;
	    }
	    s->ident = 0;
	    break;
	 case '"':
	    if (s->in_quote)
	       s->in_quote = 0;
	    else if (s->in_tag)
	    {
	       s->in_quote = 1;
	       if (bufsize == 1)
	       {
		  if (s->idname)
		     free(s->idname);
		  return NULL;
	       }
	       ptr = ++buf;
	       --bufsize;
	       /* by testing *buf against newlines, I make sure the
		  parser doesn't crash at unclosed quotes. */
	       for (*size = 0;
		    *buf != '"' && *buf != '\n' && *buf != '#' && bufsize > 0;
		    ++*size, ++buf, --bufsize);
	       /* If the quoted word contains #, it is the HTML
		  marker, in which case *size is set to actually skip
		  the unimportant part.  if it is a bg color
		  specification, it will be skipped (*size == 0) */
	       if (bufsize && *size && *buf != '\n' && s->idname
		   && idmatch(allow_mod, s->idname))
		  return ptr;
	       /* Now, if '#' had stopped the loop, get
		  on with it. */
	       if (*buf == '#' && !*size)
		  for (; *buf != '"' && *buf != '\n' && bufsize > 0;
		       ++buf, --bufsize);
	       if (bufsize == 1 || bufsize == 0)
	       {
		  if (s->idname)
		     free(s->idname);
		  return NULL;
	       }
	       if (*buf == '\n')
	       {
		  /* obviously no longer in quote. It might be well to
		     check whether '>' was encountered, but that would
		     be encouraging writers of invalid HTMLs, and we
		     don't want that, now do we? */
		  s->in_quote = 0;
		  break;
	       }
	       ++buf, --bufsize;
	       s->in_quote = 0;
	    }
	    break;
	 case ' ': case '\t': case '\n':
	    /* If a separator is found in tag, identifier is
	       finished. */
	    if (s->in_tag)
	       if (s->ident)
		  s->ident = 0;
	    break;
	 case '=': /* Ident separator + return processing */
	    if (s->in_tag)
	    {
	       s->ident = 0;
	       if (s->idname)
	       {
		  /* Added later:
		     links without quotes, like
		     <a href=http://fly.cc.fer.hr/>
		     1. skip spaces after '='
		     2. get the string (delimited by ' ', '\t',
		     '\n', '#' or '>') */
		  if (idmatch(allow_mod, s->idname))
		  {
		     /* Separator skipping */
		     for (++buf, --bufsize;
			  bufsize > 1 && strchr(" \n\t", *buf);
			  ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize <= 1)
		     {
			if (s->idname)
			   free(s->idname);
			return NULL;
		     }
		     /* If we've found a '"', give it up, and let
		      the appropriate switch deal with the bugger. */
		     if (*buf == '"')
		     {
			--buf, ++bufsize;
			break;
		     }
		     ptr = buf;
		     /* Hash support added, thanks
			to maverick@fly.cc.fer.hr */
		     for (*size = 0; bufsize > 1 && !strchr(" \n\t#>", *buf);
			  ++*size, ++buf, --bufsize)
			;
		     /* Return NULL if at EOF */
		     if (bufsize == 0)
		     {
			if (s->idname)
			   free(s->idname);
			return NULL;
		     }
		     /* Again: take care of color specifications. */
		     if (!*size)
			break;
		     /* That's it! */
		     return ptr;
		  }
	       }
	    }
	    break;
	 case '>': /* Leave the tag */
	    if (s->ident)
	    {
	       s->ident = 0;
	       if (s->idname)
	       {
		  free(s->idname);
		  s->idname = NULL;
	       }
	    }
	    s->in_tag = 0;
      }
      if (s->in_tag && (isalnum(*buf) || *buf == '/' || *buf == '_'))
      {
	 if (!s->ident)
	 {
	    if (s->idname)
	       free(s->idname);
	    s->idname = nmalloc(2);
	    s->idname[0] = *buf;
	    s->idname[1] = 0;
	    s->idlen = 1;
	    s->ident = 1;
	 }
	 else
	 {
	    ++(s->idlen);
	    s->idname = nrealloc(s->idname, s->idlen + 1);
	    s->idname[s->idlen - 1] = *buf;
	    s->idname[s->idlen] = 0;
	 }
      }
   }
   if (s->idname)
      free(s->idname);
   return NULL;
}

/* Get parser state, given a unique ID */
state_t *
getstate(char *id)
{
   static state_t *states = NULL;
   static int numstates = 0;
   int i;

   if (!numstates)
   {
      ++numstates;
      states = nmalloc(sizeof(state_t));
      init_state(states + numstates - 1, id);
      return states + numstates - 1;
   }
   else
   {
      for (i = 0; i < numstates; i++)
	 if (!strcmp(states[i].name, id))
	    break;
      if (i == numstates)
      {
	 ++numstates;
	 states = nrealloc(states, numstates * sizeof(state_t));
	 init_state(states + numstates - 1, id);
	 return states + numstates - 1;
      }
      else
	 return states + i;
   }
}

/* Initialize a parser state */
void
init_state(state_t *state, char *name)
{
   state->in_tag = state->ident = state->in_quote = state->idlen = 0;
   state->idname = NULL;
   state->name = nstrdup(name);
}

