/* main.c
   Main module, used for option parsing */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include "geturl.h"
#include "main.h"
#include "loop.h"
#include "init.h"

extern int verbose, quiet, ntry, recursive;
extern char *lfilename, *input_filename;

FILE *lfile;         /* Global log file */
#ifdef DEBUG
extern int debug;       /* Debugging on/off */
#endif /* DEBUG */

int
main (int argc, char **argv)
{
   char **url, **t;
   char *usage = "Usage: %s [options] [URL-list]\nUse -h to get help.\n";
   int i, c, nurl, status, append_to_log;
   extern int optind;
   extern char *optarg;

   append_to_log = 0;
   initialize();
   while ((c = getopt(argc, argv, "hVpqvdrHLfl:n:i:o:a:t:D:A:R:P:"))
	  != EOF)
      switch(c)
      {
	 case 'h':
	    printhelp();
	    exit(0);
	    break;
	 case 'V':
	    printf("%s\n", VERSION_STRING);
	    exit(0);
	    break;
	 case 'p':
	    setval("prefix_files", "on");
	    break;
	 case 'q':
	    setval("quiet", "on");
	    break;
	 case 'v':
	    setval("verbose", "on");
	    break;
	 case 'l':
	    setval("reclevel", optarg);
	    break;
	 case 'd':
	    setval("debug", "on");
	    break;
	 case 'n':
	    for (i = 0; i < strlen(optarg); i++)
	       switch(optarg[i])
	       {
		  case 'v':
		     setval("verbose", "off");
		     break;
		  case 'h':
		     setval("simplehostcheck", "on");
		     break;
		  case 'H':
		     setval("addhostdir", "off");
		     break;
		  case 'd':
		     setval("dirstruct", "off");
		     break;
		  case 'c':
		     setval("noclobber", "on");
		     break;
		  default:
		     printf("-n%c option is unknown. Try %s -h.\n", optarg[i],
			    argv[0]);
		     exit(1);
	       }
	    break;
	 case 'i':
	    setval("input", optarg);
	    break;
	 case 'r':
	    setval("recursive", "on");
	    break;
	 case 'H':
	    setval("spanhosts", "on");
	    break;
	 case 'L':
	    setval("relativeonly", "on");
	    break;
	 case 'f':
	    setval("followftp", "on");
	    break;
	 case 'o':
	    setval("logfile", optarg);
	    break;
	 case 'a':
	    setval("logfile", optarg);
	    append_to_log = 1;
	    break;
	 case 't':
	    setval("numtries", optarg);
	    break;
	 case 'D':
	    setval("domains", optarg);
	    break;
	 case 'A':
	    setval("accept", optarg);
	    break;
	 case 'R':
	    setval("reject", optarg);
	    break;
	 case 'P':
	    setval("dirprefix", optarg);
	    break;
	 case '?':
	    printf(usage, argv[0]);
	    exit(0);
	    break;
      }
   if (verbose == -1)
      verbose = !quiet;
   if (verbose && quiet)
   {
      printf("Can't be verbose and quiet at the same time.\n");
      printf(usage, argv[0]);
      exit(1);
   }
   /* Allocate basic pointer. */
   url = nmalloc(sizeof(char *));
   url[0] = NULL;
   /* Fill in the arguments. */
   for (nurl = 0; optind < argc; optind++, nurl++)
   {
      url = nrealloc(url, (nurl + 2) * sizeof(char *));
      url[nurl] = nstrdup(argv[optind]);
      url[nurl + 1] = NULL;
   }
   if (!url[0] && !input_filename)
   {
      printf("No URL specified.\n");
      printf(usage, argv[0]);
      exit(1);
   }
   /* Open log filename. */
   if (lfilename)
   {
      char *s;

      s = append_to_log ? "a" : "w";
      lfile = fopen(lfilename, s);
      if (lfile == NULL)
      {
	 perror("log file");
	 exit(1);
      }
   }
   else
      lfile = stdout;

   /* Call an appropriate function. */
   if (recursive)
      status = recursive_retrieve(url, input_filename, ntry, 1, NULL);
   else
      if (!input_filename)
	 status = retrieve_more_urls(url, ntry);
      else
	 status = retrieve_urls_from_file(url, input_filename, ntry);

   /* Close the log file. */
   fclose(lfile);

   /* Free the data. */
   for (t = url; *t; t++)
      free(*t);
   free(url);
   freemem();
   return !status;
}

/* Print the help message. */
void
printhelp(void)
{
   printf("---=== %s ===---\n\n", VERSION_STRING);
   printf("Options:\n");
   printf("-h           print this help message\n");
   printf("-v           verbose\n");
   printf("-q           quiet (no output)\n");
   printf("-i filename  read URL-s from file (use \"-\" for stdin)\n");
   printf("-o logfile   log messages to logfile (implies -v)\n");
   printf("-t num       set number of retries per URL (%d by default, 0 for infinity)\n", DEFAULT_NUMBER_OF_TRIES);
   printf("-nH          do not create hostname directories\n");
   printf("-r           recursive web-suck -- use with care!\n");
   printf("-l depth     maximum recursion depth (%d by default, 0 for infinity)\n", DEFAULT_MAXRECLEVEL);
   printf("-L           follow relative links only\n");
   printf("-D domnlist  comma-separated list of domains to accept\n");
   printf("-P dirname   top of directory tree\n");
   printf("-A acclist   comma-separated list of extensions to accept\n");
   printf("-R rejlist   comma-separated list of extensions to reject\n");
   printf("-f           follow FTP links from HTML documents\n");
#ifdef NO_RECURSION
   printf("note: recursion disabled!\n\n");
#endif
   printf("Refer to the manual page for a complete list of options.\n");
}

/* nmalloc, nrealloc and nstrdup exit the program if there is not
   enough memory. nstrdup also implements strdup on systems that
   do not have it. */
void *
nmalloc(size_t size)
{
   void *res;

   res = malloc(size);
   if (!res)
   {
      if (!quiet)
      {
	 fprintf(lfile, "Not enough memory.\n");
	 fflush(lfile);
      }
      exit(1);
   }
   return res;
}

void *
nrealloc(void *obj, size_t size)
{
   void *res;

   res = realloc(obj, size);
   if (!res)
   {
      if (!quiet)
      {
	 fprintf(lfile, "Not enough memory.\n");
	 fflush(lfile);
      }
      exit(1);
   }
   return res;
}

char *
nstrdup(char *s)
{
#ifdef NO_STRDUP
   int l;
   char *s1;
   
   l = strlen(s);
   s1 = nmalloc(l + 1);
   return memcpy(s1, s, l + 1);
#else
   return strdup(s);
#endif
}

