/* url.c
   Module for URL handling */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include "geturl.h"
#include "url.h"
#include "host.h"

extern int verbose, debug, prefix_files, simple_check;
extern int dirstruct, dirmode, add_hostdir, recursive;
extern char *dir_prefix;
extern FILE *lfile;

extern int errno;

/* NULL-terminated list of recognized prototypes. Note that
   recognized doesn't mean supported -- only HTTP is supported
   for now. */
char *protostrings[] = {
   "http://",
   "ftp://",
   "mailto:",
   "news:",
   "news://",
   "telnet://",
   "wais://",
   "file://",
   "gopher://",
   NULL
};

/* The same as former, but for supported protocols: */
char *sup_proto[] = {
   "http://",
   "ftp://",
   NULL
};

/* Indexes of protocols: */
uerr_t ind_proto[] = {
   URLHTTP,
   URLFTP
};

/* Index of default ports: */
uerr_t ind_ports[] = {
   DEFAULT_HTTP_PORT,
   DEFAULT_FTP_PORT
};

/* Extract the given URL of the form (http:|ftp:)//hostname(:port)?/dirname
   1. hostname (terminated with '/' or ':')
   2. port number (terminated with '/'), or chosen for the protocol
   3. dirname (everything after hostname, '/' if hostname is unterminated
   Most errors are handled. No allocation is done, you must supplie
   pointers to allocated memory */
uerr_t
parseurl(char *url, char *hostname, int *port, char *dirname)
{
   int i, l;
   uerr_t type;

   for (i = 0; sup_proto[i]; i++)
   {
      l = strlen(sup_proto[i]);
      if (!strncmp(url, sup_proto[i], l))
	 break;
   }
   if (!sup_proto[i])
      return URLUNKNOWN;
   type = ind_proto[i];
   *port = ind_ports[i];
   for (i = l; url[i] && url[i] != ':' && url[i] != '/'; i++)
      hostname[i - l] = url[i];
   hostname[i - l] = 0;
   if (url[i] == ':')
   {
      *port = 0;
      for (++i; url[i] && url[i] != '/'; i++)
	 if (isdigit(url[i]))
	 {
	    *port *= 10;
	    *port += url[i] - '0';
	 }
	 else
	 {
	    fprintf(lfile, "Bad port number.\n");
	    fflush(lfile);
	    return URLBADPORT;
	 }
   }
   if (!url[i])
      strcpy(dirname, "/");
   else
      strcpy(dirname, url + i);
   return type;
}


/* Find URL of format
   (http://|ftp://|mailto:|(...))hostname[:port]/dir -- no HTML
   parsing done, it can be any text/binary buffer */
char *
findurl(char *buf, int howmuch, int *count)
{
   char **prot, *s1, *s2;

   for (s1 = buf; howmuch; s1++, howmuch--)
      for (prot = protostrings; *prot; prot++)
	 if (howmuch <= strlen(*prot))
	    break;
	 else if (!strncasecmp(*prot, s1, strlen(*prot)))
	 {
	    for (s2 = s1, *count = 0;
		 *s2 && *s2 >= 32 && *s2 < 127 && !isspace(*s2) &&
		    !strchr(URL_SEPARATOR, *s2);
		 s2++, (*count)++);
	    return s1;
	 }
   return NULL;
}

/* 1. Parse the given URL, to extract hostname, dirname and filename
   2. Make the required hierarchy of directories, as with mkdir -p
   3. Return the pathname of the URL-equivalent filename */
char *
mkstruct(char *url)
{
   char *host, *nhost, *odir, *dir, *res;
   int port, pos, stat;
   uerr_t err;

   /* URL is like a box of chocolates -- you never know
      what you're gonna get. It's best to be safe. */
   host = nmalloc(strlen(url) + 1);
   odir = nmalloc(strlen(url) + 1);
   err = parseurl(url, host, &port, odir);
   if (err != URLHTTP && err != URLFTP)
   {
      free(odir);
      return NULL;
   }
   /* Add dir_prefix or hostname (whichever is required) to the
      beginning of dir */
   if (add_hostdir)
   {
      /* Checking for a host's true name (or at least a consistent
	 name for saving to directory) should be done here, reusing
	 the hhash, if possible. */
      if (!simple_check)
      {
	 nhost = realhost(host);
	 free(host);
	 host = nhost;
      }
      dir = nmalloc(strlen(host) + strlen(odir) + 1);
      sprintf(dir, "%s%s", host, odir);
   }
   else
   {
      dir = nmalloc(strlen(dir_prefix) + strlen(odir) + 1);
      sprintf(dir, "%s%s", dir_prefix, odir);
   }
   free(host);
   free(odir);
   for (pos = strlen(dir); pos && dir[pos] != '/'; pos--)
      if (dir[pos] == '?')
	 dir[pos] = 0; /* This takes care of CGI buggers */
   
   /* If dir is just a hostname, return index.html preceded with
      appropriate directory path. This is valid only if the URL type
      is URLHTTP. */
   if (!pos)
   {
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 free(dir);
	 return NULL;
      }
      res = nmalloc(strlen(dir) + 1 + 10 + 1);
      if (err == URLHTTP)
	 sprintf(res, "%s/index.html", dir);
      else
	 sprintf(res, "%s/", dir);
      return res;
   }
   else
   {
      dir[pos] = 0;
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 free(dir);
	 return NULL;
      }
      dir[pos] = '/';
      if (dir[pos + 1])
	 res = nstrdup(dir);
      else
      {
	 res = nmalloc(strlen(dir) + 10 + 1);
	 strcpy(res, dir);
	 if (err == URLHTTP)
	    strcat(res, "index.html");
      }
      free(dir);
      return res;
   }
}

/* The function that takes the dirname to be created, making sure that
   missing directories are made one by one. Its behaviour should be
   similar to mkdir -p. */
int
mymkdir(char *dir)
{
   int i, stat;

   /* Make a copy of dir, to be able to write to it. Otherwise, the
      function is unsafe if called with a const char *argument. */
   dir = nstrdup(dir);
   for (i = 0; 1; ++i)
   {
      for (; dir[i] && dir[i] != '/'; i++)
	 ;
      if (dir[i])
      {
	 dir[i] = 0;
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 dir[i] = '/';
      }
      else
      {
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 break;
      }
   }
   free(dir);
   return 0;
}

/* Create a unique filename, corresponding to a given URL.  If
   dirstruct is turned on, it will create all the necessary
   directories, and then return the full pathname of the file.  (it is
   important, since the function will not do what you want if you
   would like just to check for a filename) */
char *
url_filename(char *url, char *fileprefix)
{
   int count, i;
   char *name, *ptr, *murl, *pref;
   
   if (recursive && dirstruct)
   {
      char *url1;
      url1 = opt_url(url, 0);
      ptr = mkstruct(url1);
      free(url1);
      return ptr ? ptr : nstrdup(url); /* Always return *something* */
   }
   else if (prefix_files)
   {
      pref = nmalloc(strlen(dir_prefix) + 1 + strlen(fileprefix) + 1);
      sprintf(pref, "%s/%s", dir_prefix, fileprefix);
      for (count = 1; !(name = unique_name(pref, count)); count++)
	 ;
      free(pref);
      return name;
   }
   for (i = strlen(url); url[i] != '/' && i; i--)
      ;
   if (!i)
      return nstrdup(url);
   /* Checking for '~' is no longer necessary, since now it's
      done the "proper" way (through location hacks). However,
      I still provide checking for //hostname */
   if (url[i] == '/' && (!url[i + 1] || url[i - 1] == '/'))
      ptr = "index.html";
   else
      ptr = url + i + 1;
   murl = NULL;
   for (i = 0; ptr[i] && ptr[i] != '?'; i++)
      ;
   /* If there is a '?' in the string, destroy it and everything
      after it (takes care of CGI nonsense) */
   if (ptr[i] == '?')
   {
      murl = nstrdup(ptr);
      murl[i] = 0;
      ptr = murl;
   }
   pref = nmalloc(strlen(dir_prefix) + 1 + strlen(ptr) + 1);
   sprintf(pref, "%s/%s", dir_prefix, ptr);
   for (count = 0; !(name = unique_name(pref, count)); count++)
      ;
   free(pref);
   if (murl)
      free(murl);
   return name;
}

/* How many digits in an integer? */
int
numdigit(int a)
{
   int res;

   for (res = 1; a /= 10; res++);
   return res;
}

/* Return a unique filename, given a prefix and count */
char *
unique_name(char *fileprefix, int count)
{
   char *filename;

   if (count)
   {
      filename = nmalloc(strlen(fileprefix) + numdigit(count) + 2);
      sprintf(filename, "%s.%d", fileprefix, count);
   }
   else
      filename = nstrdup(fileprefix);
   if (!exists(filename))
      return filename;
   else
   {
      free(filename);
      return NULL;
   }
}

/* Does a file exist (quite bad implementation), since it will return
   that a file does not exist if, e.g., the directory is unreadable. I
   don't mind it too much currently, though.  The proper way should,
   of course, be to check errno for ENOENT, but what about missing
   directories in the path? */
int
exists(char *filename)
{
   struct stat buf;
   return stat(filename, &buf) ? 0 : 1;
}

/* Construct an absolute URL, given a (possibly) relative one.  This
   is more tricky than it might seem, but it works. */
char *
construct(char *url, char *sub, int subsize, int flag)
{
   int i, fl;
   char *constr, *t;

   t = NULL;
   if (flag)
   {
      if (*sub != '/')
      {
	 for (i = strlen(url); i && url[i] != '/'; i--);
	 if (!i || (url[i] == url[i - 1]))
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	    i = l;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i + 1);
	 constr[i + 1] = 0;
	 strncat(constr, sub, subsize);
      }
      else
      {
	 i = 0;
	 do
	 {
	    for (; url[i] && url[i] != '/'; i++);
	    if (!url[i])
	       break;
	    if ((fl = (url[i] == url[i + 1] && url[i + 1] == '/')))
	       i += 2;
	 } while (fl);
	 if (!url[i])
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i);
	 constr[i] = 0;
	 strncat(constr + i, sub, subsize);
	 constr[i + subsize] = 0;
      }
   }
   else
   {
      constr = nmalloc(subsize + 1);
      strncpy(constr, sub, subsize);
      constr[subsize] = 0;
   }
   if (t)
      free(t);
   return constr;
}

/* Remove the occurence of /anything/../ from string s.  NOTE: things
   are not that simple, because such scheme fails when confronted with
   dir/a/b/c/../../d (which should yield dir/a/d, because it tends to
   do nasty things. This was fixed by avoiding the handling of /../../
   on purpose in each pass. Thus the first pass will produce
   dir/a/b/../d, and the second dir/a/d, QEF */
char *
strip_double(char *s)
{
   char *tmp, *res, *cur, *p;
   int l, nl, ft;

   for (ft = 1;; ft = 0)
   {
      res = nmalloc(strlen(s) + 1);
      res[0] = 0;
      for (cur = s; 1 ; cur = tmp)
      {
	 if (!(tmp = strstr(cur, "/../")))
	    break;
	 strncat(res, cur, tmp - cur);
	 if (tmp[-1] == '.' && tmp[-2] == '.' && tmp[-3] == '/')
	    continue;
	 for (p = tmp - 1, l = 0; p >= cur && *p != '/'; p--, l++);
	 nl = strlen(res) - l;
	 res[nl] = 0;
	 tmp += 3;
	 while (strstr(tmp, "/../") == tmp)
	 {
	    strcat(res, "../");
	    tmp += 3;
	 }
	 ++tmp;
      }
      strcat(res, cur);
      if (!ft)
	 free(s);
      if (!strstr(res, "/../"))
	 break;
      s = res;
   }
   return res;
}

/* When /./ encountered in s, remove ./. Things are much simpler here
   than in previous example because a/././././d is always a/d (number
   of consecutive dots is not important) */
char *
strip_dot(char *s)
{
   char *tmp, *res, *cur;
   int ft;

   for (ft = 1;; ft = 0)
   {
      res = nmalloc(strlen(s) + 1);
      res[0] = 0;
      for (cur = s; 1 ; cur = tmp + 3)
      {
	 tmp = strstr(cur, "/./");
	 if (!tmp)
	    break;
	 strncat(res, cur, tmp - cur + 1);
      }
      strcat(res, cur);
      if (!ft)
	 free(s);
      else
	 ft = 0;
      if (!strstr(res, "/./"))
	 break;
      s = res;
   }
   return res;
}

/* Lousy coding, but it works... not too slow, either (speed is not
   critical here, I believe. If flag is set, call realname,
   otherwise just dot-clean the URL. */
char *
opt_url(char *url, int fl)
{
   char *s1, *s2, *host, *dir, *tmp, *res;
   int i, l, port;
   uerr_t err;
   
   s1 = strip_dot(url);
   s2 = strip_double(s1);
   free(s1);
   if (!fl)
      return s2;
   host = nmalloc(strlen(s2) + 1);
   dir = nmalloc(strlen(s2) + 1);
   err = parseurl(s2, host, &port, dir);
   if (err != URLHTTP && err != URLFTP)
   {
      free(host);
      free(dir);
      return s2;
   }
   free(s2);
   tmp = realhost(host);
   free(host);
   host = tmp;
   for (i = 0; sup_proto[i]; i++)
   {
      l = strlen(sup_proto[i]);
      if (!strncmp(url, sup_proto[i], l))
	 break;
   }
   res = nmalloc(l + strlen(host) + 1 + 6 + strlen(dir) + 1);
   sprintf(res, "%s%s:%d%s", sup_proto[i], host, port, dir);
   return res;
}

